package eu.jonahbauer.chat.bot.config;

import org.junit.jupiter.api.Test;

import java.util.List;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.OptionalLong;

import static org.junit.jupiter.api.Assertions.*;

public class BotConfigTest {

    @Test
    void buildAndAccessWithStringKeys() {
        BotConfig config = BotConfig.builder()
                .value("string", "value")
                .value("long", Long.MAX_VALUE)
                .value("double", Double.MAX_VALUE)
                .value("boolean", true)
                .value("array", List.of("foo", "bar"))
                .build();

        assertEquals(Optional.of("value"), config.getString("string"));
        assertEquals(OptionalLong.of(Long.MAX_VALUE), config.getLong("long"));
        assertEquals(OptionalDouble.of(Double.MAX_VALUE), config.getDouble("double"));
        assertEquals(Optional.of(true), config.getBoolean("boolean"));
        assertEquals(Optional.of(List.of("foo", "bar")), config.getStringArray("array"));

        assertThrows(BotConfigurationException.class, () -> config.getLong("string"));
        assertThrows(BotConfigurationException.class, () -> config.getDouble("string"));
        assertThrows(BotConfigurationException.class, () -> config.getBoolean("string"));
        assertThrows(BotConfigurationException.class, () -> config.getStringArray("string"));
    }

    @Test
    void accessUninitializedProperty() {
        var unset = BotConfig.Key.ofString("unset");
        var config = BotConfig.EMPTY;

        assertEquals(Optional.empty(), config.get(unset));
        assertEquals(Optional.empty(), config.getString("unset"));
        assertEquals(OptionalLong.empty(), config.getLong("unset"));
        assertEquals(OptionalDouble.empty(), config.getDouble("unset"));
        assertEquals(Optional.empty(), config.getBoolean("unset"));
        assertEquals(Optional.empty(), config.getStringArray("unset"));

        assertThrows(BotConfigurationException.class, () -> config.require(unset));
    }

    @Test
    void buildAndAccessWithTypedKeys() {
        var string = BotConfig.Key.ofString("string");
        var long_ = BotConfig.Key.ofLong("long");
        var double_ = BotConfig.Key.ofDouble("double");
        var boolean_ = BotConfig.Key.ofBoolean("boolean");
        var array = BotConfig.Key.ofStringArray("array");

        var config = BotConfig.builder()
                .value(string, "value")
                .value(long_, Long.MAX_VALUE)
                .value(double_, Double.MAX_VALUE)
                .value(boolean_, true)
                .value(array, List.of("foo", "bar"))
                .build();

        assertEquals(Optional.of("value"), config.getString("string"));
        assertEquals(OptionalLong.of(Long.MAX_VALUE), config.getLong("long"));
        assertEquals(OptionalDouble.of(Double.MAX_VALUE), config.getDouble("double"));
        assertEquals(Optional.of(true), config.getBoolean("boolean"));
        assertEquals(Optional.of(List.of("foo", "bar")), config.getStringArray("array"));

        assertEquals(Optional.of("value"), config.get(string));
        assertEquals(OptionalLong.of(Long.MAX_VALUE), config.get(long_));
        assertEquals(OptionalDouble.of(Double.MAX_VALUE), config.get(double_));
        assertEquals(Optional.of(true), config.get(boolean_));
        assertEquals(Optional.of(List.of("foo", "bar")), config.get(array));

        assertEquals("value", config.require(string));
        assertEquals(Long.MAX_VALUE, config.require(long_));
        assertEquals(Double.MAX_VALUE, config.require(double_));
        assertEquals(true, config.require(boolean_));
        assertEquals(List.of("foo", "bar"), config.require(array));
    }

    @Test
    void merge() {
        var a = BotConfig.builder()
                .value("a", "a")
                .value("c", "a")
                .build();

        var b = BotConfig.builder()
                .value("b", "b")
                .value("c", "b")
                .build();

        var c = a.merge(b);

        // a remains unchanged
        assertEquals(Optional.of("a"), a.getString("a"));
        assertEquals(Optional.of("a"), a.getString("c"));
        assertFalse(a.has("b"));

        // b remains unchanged
        assertEquals(Optional.of("b"), b.getString("b"));
        assertEquals(Optional.of("b"), b.getString("c"));
        assertFalse(b.has("a"));

        // c contains a's and b's values
        assertEquals(Optional.of("a"), c.getString("a"));
        assertEquals(Optional.of("b"), c.getString("b"));
        assertEquals(Optional.of("b"), c.getString("c"));
    }

}
