package eu.jonahbauer.chat.server.bot;

import eu.jonahbauer.chat.bot.api.ChatBot;
import eu.jonahbauer.chat.bot.api.Message;
import eu.jonahbauer.chat.bot.config.BotConfig;
import eu.jonahbauer.chat.bot.impl.BotCreationException;
import eu.jonahbauer.chat.bot.impl.ChatBotFactory;
import eu.jonahbauer.chat.server.Config;
import eu.jonahbauer.chat.server.management.impl.ChatBotManager;
import eu.jonahbauer.chat.server.management.impl.ChatBotSupport;
import eu.jonahbauer.chat.server.socket.SocketSupervisor;
import eu.jonahbauer.chat.server.util.Lazy;
import lombok.extern.slf4j.Slf4j;
import org.jetbrains.annotations.Blocking;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NonBlocking;
import org.jetbrains.annotations.NotNull;

import javax.management.JMException;
import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.locks.ReentrantLock;

@Slf4j
public final class ChatBotSupervisor implements AutoCloseable {
    private static final Duration POST_EXPIRATION = Duration.ofSeconds(2);
    private static final int MESSAGE_QUEUE_SIZE = 10;

    /**
     * Map of running bots indexed by their name.
     */
    private final @NotNull ConcurrentMap<@NotNull String, @NotNull RunningBot> bots = new ConcurrentHashMap<>();
    private final @NotNull ReentrantLock lock = new ReentrantLock();
    private final @NotNull Lazy<SocketSupervisor> socketSupervisor;

    public ChatBotSupervisor(@NotNull Lazy<SocketSupervisor> socketSupervisor) throws JMException {
        ChatBotManager.init(this);
        this.socketSupervisor = socketSupervisor;
    }

    /**
     * Dispatches the message to all running bots. If a bot has too many messages queued up, the message may not be
     * delivered to that bot.
     * @param message the message
     */
    public void onMessage(@NotNull Message.Post message) {
        log.debug("Dispatching message: {}", message);
        var expiration = Instant.now().plus(POST_EXPIRATION);
        for (var entry : bots.entrySet()) {
            var result = entry.getValue().offer(message, expiration);
            if (!result) log.warn("A message was dropped because bot {} is busy.", entry.getKey());
        }
    }

    /**
     * Starts all bots defined in the given config.
     * @param config a configuration
     * @throws IllegalStateException if any bots are running already
     */
    public void start(@NotNull Config config) {
        lock.lock();
        try {
            if (!bots.isEmpty()) {
                throw new IllegalStateException("start(Config) may not be used when any bots are running already");
            }
            config.bots().forEach(this::start);
        } finally {
            lock.unlock();
        }
    }

    /**
     * Starts a bot with the given name and type.
     * @param name the name of the bot (must be unique)
     * @param type the fully qualified name of a class extending {@link ChatBot}
     * @param channels the channels the bot will be active in
     * @throws BotCreationException if the bot could not be created
     */
    public void start(@NotNull String name, @NotNull String type, @NotNull List<String> channels) {
        start(name, BotConfig.builder().type(type).channels(channels).build());
    }

    /**
     * Starts a bot with the given config under the specified name.
     * @param name the name of the bot (must be unique)
     * @param config the bot configuration
     * @throws BotCreationException if the bot could not be created
     */
    public void start(@NotNull String name, @NotNull BotConfig config) {
        lock.lock();
        try {
            if (bots.containsKey(name)) {
                throw new BotCreationException("Duplicate bot name: " + name);
            }
            bots.put(name, new RunningBot(name, config));
        } finally {
            lock.unlock();
        }
    }

    /**
     * Stops the bot with the given name and waits for it to finish. Does nothing if no bot with that name exists.
     * @param name the name of the bot
     * @throws InterruptedException if any thread has interrupted the current thread.
     */
    public void stop(@NotNull String name) throws InterruptedException {
        RunningBot bot;

        lock.lock();
        try {
            bot = bots.remove(name);
            if (bot == null) return;
            bot.stop();
        } finally {
            lock.unlock();
        }

        bot.join();
    }

    /**
     * Stops all currently running bots and waits for them to finish.
     * @throws InterruptedException if any thread has interrupted the current thread.
     */
    public void stop() throws InterruptedException {
        var stopped = new ArrayList<RunningBot>();

        lock.lock();
        try {
            for (RunningBot bot : bots.values()) {
                stopped.add(bot);
                bot.stop();
            }
            bots.clear();
        } finally {
            lock.unlock();
        }

        for (var bot : stopped) {
            bot.join();
        }
    }

    @Override
    public void close() throws InterruptedException {
        stop();
    }

    /**
     * {@return a map of all currently running bots and their effective config}
     */
    public @NotNull Map<@NotNull String, @NotNull BotConfig> getBots() {
        var out = new TreeMap<String, BotConfig>();
        for (var entry : bots.entrySet()) {
            out.put(entry.getKey(), entry.getValue().getConfig());
        }
        return Collections.unmodifiableMap(out);
    }

    @Contract("_ -> new")
    @SuppressWarnings("unchecked")
    private static @NotNull ChatBotFactory<?> getChatBotFactory(@NotNull String type) {
        try {
            var clazz = Class.forName(type);
            if (!ChatBot.class.isAssignableFrom(clazz)) {
                throw new BotCreationException("Not a chat bot type: " + type);
            }

            return new ChatBotFactory<>((Class<? extends ChatBot>) clazz);
        } catch (ClassNotFoundException ex) {
            throw new BotCreationException("Unknown chat bot of type: " + type, ex);
        }
    }

    /**
     * Manages a {@link ChatBot} instance running on its own thread. Takes care of recreating the bot after it threw
     * an exception. {@linkplain ChatBotSupport#register(ChatBotSupervisor, String, BotConfig) Registers} the bot
     * during construction and {@linkplain ChatBotSupport#unregister(String) unregisters} it during {@link #stop()}.
     */
    private class RunningBot implements Runnable {
        private final @NotNull String name;
        private final @NotNull ChatBot bot;

        private final @NotNull Thread thread;
        private final @NotNull BlockingQueue<@NotNull PendingPost> queue = new ArrayBlockingQueue<>(MESSAGE_QUEUE_SIZE);
        private final @NotNull AtomicBoolean stopped = new AtomicBoolean();

        public RunningBot(@NotNull String name, @NotNull BotConfig config) {
            this.name = name;

            var factory = getChatBotFactory(config.getType());
            this.bot = factory.create(config);

            log.info("starting bot {}...", name);
            ChatBotSupport.register(ChatBotSupervisor.this, name, config);
            this.thread = Thread.ofVirtual().name("ChatBot[" + name + "]").start(this);
        }

        @Override
        public void run() {
            log.info("started bot {}", name);
            while (!stopped.get()) {
                try {
                    var post = queue.take();
                    if (Instant.now().isAfter(post.expiration())) {
                        log.warn("A message was dropped because it expired: {}", post);
                    } else if (bot.getConfig().isIgnoreBots() && post.post().bot()) {
                        log.debug("Ignoring message {} because bottag is set.", post.post());
                    } else if (!bot.getConfig().getChannels().contains(post.post().channel())) {
                        log.debug("Ignoring message {} because channel is not listened to.", post.post());
                    } else {
                        bot.onMessage(ChatBotSupervisor.this.socketSupervisor.get(), post.post());
                    }
                } catch (InterruptedException _) {
                    // ignore
                } catch (Exception ex) {
                    log.warn("bot {} threw an exception during message handling", name, ex);
                } catch (Throwable ex) {
                    log.error("bot {} threw an error during message handling", name, ex);
                }
            }

            try {
                bot.onStop();
            } catch (Exception ex) {
                log.warn("bot {} threw an exception during shutdown", name, ex);
            }

            log.info("stopped bot {}", name);
        }

        @NonBlocking
        public boolean offer(@NotNull Message.Post post, @NotNull Instant expiration) {
            if (stopped.get()) return false;
            return queue.offer(new PendingPost(post, expiration));
        }

        /**
         * Stops the bot, unregistering it from the MBean server and preventing it from receiving any new messages.
         */
        @NonBlocking
        public void stop() {
            if (stopped.getAndSet(true)) return;
            log.info("stopping bot {}", name);
            ChatBotSupport.unregister(name);
            thread.interrupt();
        }

        /**
         * Waits for the bot to finish execution.
         */
        @Blocking
        public void join() throws InterruptedException {
            thread.join();
        }

        public @NotNull BotConfig getConfig() {
            return bot.getConfig();
        }

        private record PendingPost(@NotNull Message.Post post, @NotNull Instant expiration) {}
    }
}
