package eu.jonahbauer.chat.server.socket;

import com.google.gson.*;
import eu.jonahbauer.chat.bot.api.Chat;
import eu.jonahbauer.chat.bot.api.Message;
import eu.jonahbauer.chat.server.Config;
import eu.jonahbauer.chat.server.bot.ChatBotSupervisor;
import eu.jonahbauer.chat.server.json.BooleanTypeAdapter;
import eu.jonahbauer.chat.server.json.LocalDateTimeTypeAdapter;
import eu.jonahbauer.chat.server.json.RuntimeTypeAdapterFactory;
import eu.jonahbauer.chat.server.management.SocketState;
import eu.jonahbauer.chat.server.management.impl.SocketManager;
import eu.jonahbauer.chat.server.management.impl.SocketSupport;
import eu.jonahbauer.chat.server.util.Lazy;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.management.JMException;
import java.net.CookieManager;
import java.net.URI;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpRequest.BodyPublishers;
import java.net.http.HttpResponse;
import java.net.http.WebSocket;
import java.nio.ByteBuffer;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.locks.ReentrantLock;

import static eu.jonahbauer.chat.server.util.UrlTemplateProcessor.URL;

@Slf4j
public final class SocketSupervisor implements Chat, AutoCloseable {
    private static final URI AUTH_SERVER = URI.create("https://chat.qed-verein.de/rubychat/account");
    private static final String ORIGIN = "https://chat.qed-verein.de";
    private static final String SERVER = "wss://chat.qed-verein.de/websocket?position=0&version=2&channel=";
    private static final int PING_INTERVAL = 30;

    private static final ScheduledExecutorService SCHEDULER = Executors.newScheduledThreadPool(1, Thread.ofVirtual().name("SocketSupervisor").factory());
    private static final Gson GSON = new GsonBuilder()
            .registerTypeAdapterFactory(RuntimeTypeAdapterFactory.of(Message.class)
                    .registerSubtype(Message.Ping.class, "ping")
                    .registerSubtype(Message.Post.class, "post")
                    .registerSubtype(Message.Ack.class, "ack")
                    .registerSubtype(Message.Pong.class, "pong")
            )
            .registerTypeAdapter(boolean.class, new BooleanTypeAdapter())
            .registerTypeAdapter(Boolean.class, new BooleanTypeAdapter())
            .registerTypeAdapter(LocalDateTime.class, new LocalDateTimeTypeAdapter())
            .create();

    private volatile @Nullable Config.Account account;
    private volatile @Nullable Credentials credentials;

    private final @NotNull CookieManager cookie = new CookieManager();
    private final @NotNull HttpClient client = HttpClient.newBuilder().cookieHandler(cookie).build();
    private final @NotNull ConcurrentMap<@NotNull String, @NotNull ChatClient> sockets = new ConcurrentHashMap<>();
    private final @NotNull Lazy<ChatBotSupervisor> chatBotSupervisor;
    private final @NotNull ReentrantLock lock = new ReentrantLock();

    public SocketSupervisor(@NotNull Lazy<ChatBotSupervisor> chatBotSupervisor) throws JMException {
        SocketManager.init(this);
        this.chatBotSupervisor = chatBotSupervisor;
    }

    public void setAccount(@NotNull Config.Account account) {
        this.account = Objects.requireNonNull(account, "account");
    }

    public void setAccount(@NotNull String username, @NotNull String password) {
        this.account = new Config.Account(username, password);
    }

    @Override
    public boolean send(@NotNull String channel, @NotNull String name, @NotNull String message, boolean bottag, boolean publicId) {
        var socket = sockets.get(channel);
        if (socket == null || !socket.isOpen()) {
            log.error("Cannot deliver message to {}: not connected", channel);
            return false;
        } else {
            var out = new OutgoingMessage(name, message, channel, socket.delay, publicId, bottag);
            try {
                socket.send(GSON.toJson(out, OutgoingMessage.class));
                log.info("Sending message: {}", out);
                return true;
            } catch (Exception e) {
                log.error("Could not serialize message: {}", out, e);
                return false;
            }
        }
    }

    /**
     * Starts all sockets defined in the given config.
     * @param config a configuration
     * @throws IllegalStateException if any sockets are already running
     */
    public void start(@NotNull Config config) {
        lock.lock();
        try {
            if (!this.sockets.isEmpty()) {
                throw new IllegalStateException("start(Config) may not be used when any sockets are running already");
            }

            setAccount(config.account());
            config.channels().forEach(this::start);
        } finally {
            lock.unlock();
        }
    }

    /**
     * Starts a socket for the given channel.
     * @param channel the channel
     * @throws IllegalStateException if a socket is already connected to that channel
     */
    public void start(@NotNull String channel) {
        lock.lock();
        try {
            if (sockets.containsKey(channel)) {
                throw new SocketCreationException("Duplicate channel: " + channel);
            }

            var socket = new ChatClient(channel);
            this.sockets.put(channel, socket);
        } finally {
            lock.unlock();
        }
    }

    /**
     * Forcefully restarts the socket for the given channel when it is on cooldown.
     * @param channel the channel
     * @throws IllegalStateException if the socket is not on cooldown
     */
    public void restart(@NotNull String channel) {
        var socket = sockets.get(channel);
        if (socket != null) socket.restart();
    }

    public @Nullable SocketState getState(@NotNull String channel) {
        var socket = sockets.get(channel);
        return socket == null ? null : socket.getState();

    }

    public @Nullable Instant getCooldownUntil(@NotNull String channel) {
        var socket = sockets.get(channel);
        return socket == null ? null : socket.getCooldownUntil();
    }

    /**
     * Stops the socket for the given channel and waits for it to finish.
     * @param channel the channel
     * @throws InterruptedException if any thread has interrupted the current thread.
     */
    public void stop(@NotNull String channel) throws InterruptedException {
        var socket = sockets.get(channel);
        if (socket == null) return;

        socket.stop();
        socket.join();
    }

    /**
     * Stops all currently running sockets and waits for them to finish.
     * @throws InterruptedException if any thread has interrupted the current thread.
     */
    public void stop() throws InterruptedException {
        var sockets = new ArrayList<>(this.sockets.values());

        for (var socket : sockets) {
            socket.stop();
        }

        for (var socket : sockets) {
            socket.join();
        }
    }

    @Override
    public void close() throws InterruptedException {
        stop();
    }

    public @NotNull SortedSet<@NotNull String> getChannels() {
        return Collections.unmodifiableSortedSet(new TreeSet<>(sockets.keySet()));
    }

    @SneakyThrows
    private @NotNull Credentials login() {
        var account = this.account;
        if (account == null) throw new IllegalStateException("Account not initialized.");

        var credentials = this.credentials;
        if (credentials != null) return credentials;

        var request = HttpRequest.newBuilder().uri(AUTH_SERVER)
                .POST(BodyPublishers.ofString(
                        URL."username=\{account.username()}&password=\{account.password()}&version=20171030131648"
                ))
                .build();
        var response = client.send(request, HttpResponse.BodyHandlers.ofString());

        if (response.statusCode() != 200) {
            throw new IllegalArgumentException(STR."invalid credentials (status code: \{response.statusCode()})");
        }

        var body = response.body();
        if (!body.contains("success")) {
            throw new IllegalArgumentException(STR."invalid credentials (\{body})");
        }

        var cookies = cookie.getCookieStore().get(AUTH_SERVER);
        var userid = cookies.stream().filter(cookie -> cookie.getName().equals("userid")).findFirst();
        var pwhash = cookies.stream().filter(cookie -> cookie.getName().equals("pwhash")).findFirst();

        if (userid.isEmpty() || pwhash.isEmpty()) {
            throw new IllegalArgumentException(STR."invalid credentials (status code: \{response.statusCode()})");
        }

        this.credentials = credentials = new Credentials(userid.get().getValue(), pwhash.get().getValue());
        return credentials;
    }

    private record Credentials(@NotNull String userid, @NotNull String pwhash) {}


    private sealed interface ChatClientState {

        default void onEnter() {}

        default void send(@NotNull String message) {
            throw new IllegalStateException();
        }

        default void restart() {
            throw new IllegalStateException();
        }

        default void stop() {
            abort();
        }

        default void abort() {
            throw new IllegalStateException();
        }

        @NotNull SocketState getState();
    }

    private final class ChatClient implements WebSocket.Listener, ChatClientState {
        private final @NotNull String channel;
        private final @NotNull ReentrantLock lock = new ReentrantLock();
        private final @NotNull CountDownLatch stopped = new CountDownLatch(1);

        private @NotNull ChatClientState state = new Created();
        private volatile long delay = -1;

        public ChatClient(@NotNull String channel) {
            this.channel = channel;
            SocketSupport.register(SocketSupervisor.this, channel);
            this.state.onEnter();
        }

        private void transition(@NotNull ChatClientState from, @NotNull ChatClientState to) {
            lock.lock();
            try {
                if (state == from) {
                    state = to;
                    state.onEnter();
                }
            } finally {
                lock.unlock();
            }
        }

        public boolean isOpen() {
            return getState() == SocketState.CONNECTED;
        }

        public @Nullable Instant getCooldownUntil() {
            return state instanceof Cooldown cooldown ? cooldown.until : null;
        }

        @Override
        public void send(@NotNull String message) {
            lock.lock();
            try {
                state.send(message);
            } finally {
                lock.unlock();
            }
        }

        @Override
        public void restart() {
            lock.lock();
            try {
                state.restart();
            } finally {
                lock.unlock();
            }
        }

        @Override
        public void stop() {
            lock.lock();
            try {
                state.stop();
            } finally {
                lock.unlock();
            }
        }

        @Override
        public void abort() {
            lock.lock();
            try {
                state.abort();
            } finally {
                lock.unlock();
            }
        }

        @Override
        public @NotNull SocketState getState() {
            return state.getState();
        }

        public void join() throws InterruptedException {
            stopped.await();
        }

        private <T extends Throwable> void withNamedThread(@NotNull ThrowingRunnable<T> runnable) throws T {
            var name = Thread.currentThread().getName();
            try {
                Thread.currentThread().setName(STR."Socket[\{channel}]");
                runnable.run();
            } finally {
                Thread.currentThread().setName(name);
            }
        }

        private <R, T extends Throwable> R withNamedThread(@NotNull ThrowingSupplier<R, T> supplier) throws T {
            var name = Thread.currentThread().getName();
            try {
                Thread.currentThread().setName(STR."Socket[\{channel}]");
                return supplier.get();
            } finally {
                Thread.currentThread().setName(name);
            }
        }

        private interface ThrowingRunnable<T extends Throwable> {
            void run() throws T;
        }

        private interface ThrowingSupplier<R, T extends Throwable> {
            R get() throws T;
        }

        public final class Created implements ChatClientState {
            @Override
            public void onEnter() {
                transition(this, new Connecting());
            }

            @Override
            public @NotNull SocketState getState() {
                return SocketState.CREATED;
            }
        }

        public final class Connecting implements ChatClientState {
            private final int cooldown;

            public Connecting() {
                this(1);
            }

            public Connecting(int cooldown) {
                this.cooldown = cooldown;
            }

            @Override
            public void onEnter() {
                log.info("starting socket {}", channel);
                try {
                    var connected = new Connected();
                    transition(this, connected);
                } catch (Exception ex) {
                    log.warn("socket {} failed with an exception.", channel, ex);
                    transition(this, new Cooldown(cooldown));
                }
            }

            @Override
            public @NotNull SocketState getState() {
                return SocketState.CONNECTING;
            }
        }

        public final class Connected implements ChatClientState, WebSocket.Listener {
            private final @NotNull WebSocket socket;
            private @Nullable ScheduledFuture<?> ping;

            private volatile boolean stopped;

            private final @NotNull List<@NotNull CharSequence> parts = new ArrayList<>();
            private volatile CompletableFuture<?> message = new CompletableFuture<>();

            public Connected() throws ExecutionException, InterruptedException {
                var credentials = SocketSupervisor.this.login();
                this.socket = SocketSupervisor.this.client.newWebSocketBuilder()
                        .header("Origin", ORIGIN)
                        .header("Cookie", URL."userid=\{credentials.userid()}; pwhash=\{credentials.pwhash()}")
                        .buildAsync(URI.create(SERVER + URL."\{channel}"), this)
                        .get();
            }

            @Override
            public void onOpen(@NotNull WebSocket webSocket) {
                withNamedThread(() -> {
                    log.info("started socket {}", channel);
                    ping = SCHEDULER.scheduleAtFixedRate(this::ping, PING_INTERVAL, PING_INTERVAL, TimeUnit.SECONDS);

                    WebSocket.Listener.super.onOpen(webSocket);
                });
            }

            @Override
            public @Nullable CompletionStage<?> onText(@NotNull WebSocket webSocket, @NotNull CharSequence message, boolean last) {
                return withNamedThread(() -> {
                    this.parts.add(message);
                    webSocket.request(1);

                    if (last) {
                        onMessage(this.parts);
                        this.parts.clear();
                        this.message.complete(null);

                        CompletionStage<?> out = this.message;
                        this.message = new CompletableFuture<>();
                        return out;
                    }

                    return this.message;
                });
            }

            private void onMessage(@NotNull List<@NotNull CharSequence> parts) {
                var text = String.join("", parts);
                Message message;
                try {
                    message = GSON.fromJson(text, Message.class);
                } catch (Exception e) {
                    log.warn("Could not parse as message: {}", text, e);
                    return;
                }

                if (message instanceof Message.Post) {
                    log.info("Received message: {}", message);
                } else {
                    log.debug("Received message: {}", message);
                }

                if (message instanceof Message.Post post) {
                    delay = post.id() + 1;
                    SocketSupervisor.this.chatBotSupervisor.get().onMessage(post);
                }
            }

            @Override
            public @Nullable CompletionStage<?> onBinary(WebSocket webSocket, ByteBuffer data, boolean last) {
                return withNamedThread(() -> {
                    log.warn("Socket {} received binary data.", channel);
                    return WebSocket.Listener.super.onBinary(webSocket, data, last);
                });
            }

            @Override
            public @Nullable CompletionStage<?> onClose(@NotNull WebSocket webSocket, int statusCode, @NotNull String reason) {
                return withNamedThread(() -> {
                    log.info("socket {} closed (code={}, reason={})", channel, statusCode, reason);

                    lock.lock();
                    try {
                        if (stopped) {
                            transition(this, new Stopped());
                        } else {
                            transition(this, new Cooldown(1));
                        }
                    } finally {
                        lock.unlock();
                    }
                    return null;
                });
            }

            @Override
            public void onError(WebSocket webSocket, Throwable error) {
                withNamedThread(() -> {
                    log.warn("Socket {} failed with an exception.", channel, error);

                    lock.lock();
                    try {
                        if (stopped) {
                            transition(this, new Stopped());
                        } else {
                            transition(this, new Cooldown(1));
                        }
                    } finally {
                        lock.unlock();
                    }
                });
            }

            @Override
            public void send(@NotNull String message) {
                if (stopped) throw new IllegalStateException();
                socket.sendText(message, true);
            }

            @Override
            public void abort() {
                if (ping != null) ping.cancel(true);
                socket.abort();
                transition(this, new Stopped());
            }

            @Override
            public void stop() {
                if (stopped) throw new IllegalStateException();
                log.info("stopping socket {}", channel);
                if (ping != null) ping.cancel(true);
                stopped = true;
                socket.sendClose(WebSocket.NORMAL_CLOSURE, "ok");
            }

            private void ping() {
                socket.sendText(GSON.toJson(Message.PING, Message.class), true);
                log.debug("Sending message: {}", Message.PING);
            }

            @Override
            public @NotNull SocketState getState() {
                return stopped ? SocketState.STOPPING : SocketState.CONNECTED;
            }
        }

        public final class Cooldown implements ChatClientState {
            private final int cooldown;

            private Instant until;
            private ScheduledFuture<?> future;

            public Cooldown(int cooldown) {
                this.cooldown = cooldown;
            }

            @Override
            public void onEnter() {
                log.info("restarting socket {} in {} seconds", channel, cooldown);

                var nextCooldown = Math.min(3600, cooldown * 2);
                this.until = Instant.now().plusSeconds(cooldown);
                this.future = SCHEDULER.schedule(() -> transition(this, new Connecting(nextCooldown)), cooldown, TimeUnit.SECONDS);
            }

            @Override
            public void abort() {
                log.info("stopping socket {}", channel);
                future.cancel(true);
                transition(this, new Stopped());
            }

            @Override
            public void restart() {
                future.cancel(true);
                transition(this, new Connecting());
            }

            @Override
            public @NotNull SocketState getState() {
                return SocketState.COOLDOWN;
            }
        }

        public final class Stopped implements ChatClientState {

            @Override
            public void onEnter() {
                log.info("stopped socket {}", channel);
                stopped.countDown();
                sockets.remove(channel, ChatClient.this);
                SocketSupport.unregister(channel);
            }

            @Override
            public void abort() {}

            @Override
            public @NotNull SocketState getState() {
                return SocketState.STOPPED;
            }
        }
    }
}
